@@ -15,6 +15,7 @@
 import contextlib
 import errno
 import hashlib
+import importlib
 import logging
 import math
 import os
@@ -37,7 +38,7 @@ import glance_store.location
 
 try:
     from cinderclient import exceptions as cinder_exception
-    from cinderclient.v2 import client as cinderclient
+    from cinderclient.v3 import client as cinderclient
     from os_brick.initiator import connector
     from oslo_privsep import priv_context
 except ImportError:
@@ -304,84 +305,47 @@ Possible values:
 Related options:
     * None
 
+"""),
+    cfg.BoolOpt('cinder_enforce_multipath',
+                default=False,
+                help="""
+If this is set to True, attachment of volumes for image transfer will
+be aborted when multipathd is not running. Otherwise, it will fallback
+to single path.
+
+Possible values:
+    * True or False
+
+Related options:
+    * cinder_use_multipath
+
+"""),
+    cfg.BoolOpt('cinder_use_multipath',
+                default=False,
+                help="""
+Flag to identify mutipath is supported or not in the deployment.
+
+Set it to False if multipath is not supported.
+
+Possible values:
+    * True or False
+
+Related options:
+    * cinder_enforce_multipath
+
+"""),
+    cfg.StrOpt('cinder_mount_point_base',
+               default='/var/lib/glance/mnt',
+               help="""
+Directory where the NFS volume is mounted on the glance node.
+
+Possible values:
+
+* A string representing absolute path of mount point.
 """),
 ]
 
 
-def get_root_helper(backend=None):
-    if backend:
-        rootwrap = getattr(CONF, backend).rootwrap_config
-    else:
-        rootwrap = CONF.glance_store.rootwrap_config
-
-    return 'sudo glance-rootwrap %s' % rootwrap
-
-
-def is_user_overriden(conf, backend=None):
-    if backend:
-        store_conf = getattr(conf, backend)
-    else:
-        store_conf = conf.glance_store
-
-    return all([store_conf.get('cinder_store_' + key)
-                for key in ['user_name', 'password',
-                            'project_name', 'auth_address']])
-
-
-def get_cinderclient(conf, context=None, backend=None):
-    if backend:
-        glance_store = getattr(conf, backend)
-    else:
-        glance_store = conf.glance_store
-
-    user_overriden = is_user_overriden(conf, backend=backend)
-    if user_overriden:
-        username = glance_store.cinder_store_user_name
-        password = glance_store.cinder_store_password
-        project = glance_store.cinder_store_project_name
-        url = glance_store.cinder_store_auth_address
-    else:
-        username = context.user
-        password = context.auth_token
-        project = context.tenant
-
-        if glance_store.cinder_endpoint_template:
-            url = glance_store.cinder_endpoint_template % context.to_dict()
-        else:
-            info = glance_store.cinder_catalog_info
-            service_type, service_name, interface = info.split(':')
-            try:
-                catalog = keystone_sc.ServiceCatalogV2(context.service_catalog)
-                url = catalog.url_for(
-                    region_name=glance_store.cinder_os_region_name,
-                    service_type=service_type,
-                    service_name=service_name,
-                    interface=interface)
-            except keystone_exc.EndpointNotFound:
-                reason = _("Failed to find Cinder from a service catalog.")
-                raise exceptions.BadStoreConfiguration(store_name="cinder",
-                                                       reason=reason)
-
-    c = cinderclient.Client(username,
-                            password,
-                            project,
-                            auth_url=url,
-                            region_name=glance_store.cinder_os_region_name,
-                            insecure=glance_store.cinder_api_insecure,
-                            retries=glance_store.cinder_http_retries,
-                            cacert=glance_store.cinder_ca_certificates_file)
-
-    LOG.debug('Cinderclient connection created for user %(user)s using URL: '
-              '%(url)s.', {'user': username, 'url': url})
-
-    # noauth extracts user_id:project_id from auth_token
-    if not user_overriden:
-        c.client.auth_token = context.auth_token or '%s:%s' % (username,
-                                                               project)
-    c.client.management_url = url
-    return c
-
-
 class StoreLocation(glance_store.location.StoreLocation):
 
     """Class describing a Cinder URI."""
@@ -405,24 +369,6 @@ class StoreLocation(glance_store.location.StoreLocation):
             raise exceptions.BadStoreUri(message=reason)
 
 
-@contextlib.contextmanager
-def temporary_chown(path, backend=None):
-    owner_uid = os.getuid()
-    orig_uid = os.stat(path).st_uid
-
-    if orig_uid != owner_uid:
-        processutils.execute('chown', owner_uid, path,
-                             run_as_root=True,
-                             root_helper=get_root_helper(backend=backend))
-    try:
-        yield
-    finally:
-        if orig_uid != owner_uid:
-            processutils.execute('chown', orig_uid, path,
-                                 run_as_root=True,
-                                 root_helper=get_root_helper(backend=backend))
-
-
 class Store(glance_store.driver.Store):
 
     """Cinder backend store adapter."""
@@ -437,6 +383,99 @@ class Store(glance_store.driver.Store):
         super(Store, self).__init__(*args, **kargs)
         if self.backend_group:
             self._set_url_prefix()
+        # We are importing it here to let the config options load
+        # before we use them in the fs_mount file
+        self.mount = importlib.import_module('glance_store.common.fs_mount')
+
+    def get_root_helper(self):
+        if self.backend_group:
+            rootwrap = getattr(CONF, self.backend_group).rootwrap_config
+        else:
+            rootwrap = CONF.glance_store.rootwrap_config
+
+        return 'sudo glance-rootwrap %s' % rootwrap
+
+    def is_user_overriden(self):
+        if self.backend_group:
+            store_conf = getattr(self.conf, self.backend_group)
+        else:
+            store_conf = self.conf.glance_store
+
+        return all([store_conf.get('cinder_store_' + key)
+                    for key in ['user_name', 'password',
+                                'project_name', 'auth_address']])
+
+    def get_cinderclient(self, context=None):
+        if self.backend_group:
+            glance_store = getattr(self.conf, self.backend_group)
+        else:
+            glance_store = self.conf.glance_store
+
+        user_overriden = self.is_user_overriden()
+        if user_overriden:
+            username = glance_store.cinder_store_user_name
+            password = glance_store.cinder_store_password
+            project = glance_store.cinder_store_project_name
+            url = glance_store.cinder_store_auth_address
+        else:
+            username = context.user
+            password = context.auth_token
+            project = context.tenant
+
+            if glance_store.cinder_endpoint_template:
+                url = glance_store.cinder_endpoint_template % context.to_dict()
+            else:
+                info = glance_store.cinder_catalog_info
+                service_type, service_name, interface = info.split(':')
+                try:
+                    catalog = keystone_sc.ServiceCatalogV2(
+                        context.service_catalog)
+                    url = catalog.url_for(
+                        region_name=glance_store.cinder_os_region_name,
+                        service_type=service_type,
+                        service_name=service_name,
+                        interface=interface)
+                except keystone_exc.EndpointNotFound:
+                    reason = _("Failed to find Cinder from a service catalog.")
+                    raise exceptions.BadStoreConfiguration(store_name="cinder",
+                                                           reason=reason)
+
+        c = cinderclient.Client(
+            username, password, project, auth_url=url,
+            region_name=glance_store.cinder_os_region_name,
+            insecure=glance_store.cinder_api_insecure,
+            retries=glance_store.cinder_http_retries,
+            cacert=glance_store.cinder_ca_certificates_file)
+
+        LOG.debug(
+            'Cinderclient connection created for user %(user)s using URL: '
+            '%(url)s.', {'user': username, 'url': url})
+
+        # noauth extracts user_id:project_id from auth_token
+        if not user_overriden:
+            c.client.auth_token = context.auth_token or '%s:%s' % (username,
+                                                                   project)
+        c.client.management_url = url
+        return c
+
+    @contextlib.contextmanager
+    def temporary_chown(self, path):
+        owner_uid = os.getuid()
+        orig_uid = os.stat(path).st_uid
+
+        if orig_uid != owner_uid:
+            processutils.execute(
+                'chown', owner_uid, path,
+                run_as_root=True,
+                root_helper=self.get_root_helper())
+        try:
+            yield
+        finally:
+            if orig_uid != owner_uid:
+                processutils.execute(
+                    'chown', orig_uid, path,
+                    run_as_root=True,
+                    root_helper=self.get_root_helper())
 
     def get_schemes(self):
         return ('cinder',)
@@ -445,8 +484,7 @@ class Store(glance_store.driver.Store):
         self._url_prefix = "cinder://"
 
     def _check_context(self, context, require_tenant=False):
-        user_overriden = is_user_overriden(self.conf,
-                                           backend=self.backend_group)
+        user_overriden = self.is_user_overriden()
         if user_overriden and not require_tenant:
             return
         if context is None:
@@ -491,15 +529,43 @@ class Store(glance_store.driver.Store):
             raise exceptions.BackendException(msg)
         return volume
 
+    def get_hash_str(self, base_str):
+        """Returns string that represents SHA256 hash of base_str (in hex format).
+
+        If base_str is a Unicode string, encode it to UTF-8.
+        """
+        if isinstance(base_str, str):
+            base_str = base_str.encode('utf-8')
+        return hashlib.sha256(base_str).hexdigest()
+
+    def _get_mount_path(self, share, mount_point_base):
+        """Returns the mount path prefix using the mount point base and share.
+
+        :returns: The mount path prefix.
+        """
+        return os.path.join(mount_point_base, self.get_hash_str(share))
+
     @contextlib.contextmanager
     def _open_cinder_volume(self, client, volume, mode):
         attach_mode = 'rw' if mode == 'wb' else 'ro'
         device = None
-        root_helper = get_root_helper(backend=self.backend_group)
+        root_helper = self.get_root_helper()
         priv_context.init(root_helper=shlex.split(root_helper))
         host = socket.gethostname()
-        properties = connector.get_connector_properties(root_helper, host,
-                                                        False, False)
+        if self.backend_group:
+            use_multipath = getattr(
+                self.conf, self.backend_group).cinder_use_multipath
+            enforce_multipath = getattr(
+                self.conf, self.backend_group).cinder_enforce_multipath
+            mount_point_base = getattr(
+                self.conf, self.backend_group).cinder_mount_point_base
+        else:
+            use_multipath = self.conf.glance_store.cinder_use_multipath
+            enforce_multipath = self.conf.glance_store.cinder_enforce_multipath
+            mount_point_base = self.conf.glance_store.cinder_mount_point_base
+
+        properties = connector.get_connector_properties(
+            root_helper, host, use_multipath, enforce_multipath)
 
         try:
             volume.reserve(volume)
@@ -514,16 +580,31 @@ class Store(glance_store.driver.Store):
             conn = connector.InitiatorConnector.factory(
                 connection_info['driver_volume_type'], root_helper,
                 conn=connection_info)
-            device = conn.connect_volume(connection_info['data'])
+            if connection_info['driver_volume_type'] == 'nfs':
+                @utils.synchronized(connection_info['data']['export'])
+                def connect_volume_nfs():
+                    data = connection_info['data']
+                    export = data['export']
+                    vol_name = data['name']
+                    mountpoint = self._get_mount_path(
+                        export,
+                        os.path.join(mount_point_base, 'nfs'))
+                    options = data['options']
+                    self.mount.mount(
+                        'nfs', export, vol_name, mountpoint, host,
+                        root_helper, options)
+                    return {'path': os.path.join(mountpoint, vol_name)}
+                device = connect_volume_nfs()
+            else:
+                device = conn.connect_volume(connection_info['data'])
             volume.attach(None, 'glance_store', attach_mode, host_name=host)
             volume = self._wait_volume_status(volume, 'attaching', 'in-use')
             if (connection_info['driver_volume_type'] == 'rbd' and
                not conn.do_local_attach):
                 yield device['path']
             else:
-                with temporary_chown(device['path'],
-                                     backend=self.backend_group), \
-                        open(device['path'], mode) as f:
+                with self.temporary_chown(
+                        device['path']), open(device['path'], mode) as f:
                     yield f
         except Exception:
             LOG.exception(_LE('Exception while accessing to cinder volume '
@@ -537,7 +618,15 @@ class Store(glance_store.driver.Store):
 
             if device:
                 try:
-                    conn.disconnect_volume(connection_info['data'], device)
+                    if connection_info['driver_volume_type'] == 'nfs':
+                        @utils.synchronized(connection_info['data']['export'])
+                        def disconnect_volume_nfs():
+                            path, vol_name = device['path'].rsplit('/', 1)
+                            self.mount.umount(vol_name, path, host,
+                                              root_helper)
+                        disconnect_volume_nfs()
+                    else:
+                        conn.disconnect_volume(connection_info['data'], device)
                 except Exception:
                     LOG.exception(_LE('Failed to disconnect volume '
                                       '%(volume_id)s.'),
@@ -600,8 +689,7 @@ class Store(glance_store.driver.Store):
         loc = location.store_location
         self._check_context(context)
         try:
-            client = get_cinderclient(self.conf, context,
-                                      backend=self.backend_group)
+            client = self.get_cinderclient(context)
             volume = client.volumes.get(loc.volume_id)
             size = int(volume.metadata.get('image_size',
                                            volume.size * units.Gi))
@@ -635,9 +723,7 @@ class Store(glance_store.driver.Store):
 
         try:
             self._check_context(context)
-            volume = get_cinderclient(
-                self.conf, context,
-                backend=self.backend_group).volumes.get(loc.volume_id)
+            volume = self.get_cinderclient(context).volumes.get(loc.volume_id)
             return int(volume.metadata.get('image_size',
                                            volume.size * units.Gi))
         except cinder_exception.NotFound:
@@ -671,8 +757,7 @@ class Store(glance_store.driver.Store):
         """
 
         self._check_context(context, require_tenant=True)
-        client = get_cinderclient(self.conf, context,
-                                  backend=self.backend_group)
+        client = self.get_cinderclient(context)
         os_hash_value = hashlib.new(str(hashing_algo))
         checksum = hashlib.md5()
         bytes_written = 0
@@ -774,7 +859,7 @@ class Store(glance_store.driver.Store):
 
         image_metadata = {}
         if self.backend_group:
-            image_metadata['backend'] = u"%s" % self.backend_group
+            image_metadata['store'] = u"%s" % self.backend_group
 
         return ('cinder://%s' % volume.id,
                 bytes_written,
@@ -797,9 +882,7 @@ class Store(glance_store.driver.Store):
         loc = location.store_location
         self._check_context(context)
         try:
-            volume = get_cinderclient(
-                self.conf, context,
-                backend=self.backend_group).volumes.get(loc.volume_id)
+            volume = self.get_cinderclient(context).volumes.get(loc.volume_id)
             volume.delete()
         except cinder_exception.NotFound:
             raise exceptions.NotFound(image=loc.volume_id)
