@@ -16,8 +16,9 @@
 import contextlib
 import errno
 import hashlib
-import mock
 import os
+from unittest import mock
+
 import six
 import socket
 import sys
@@ -36,7 +37,7 @@ from glance_store.tests import base
 from glance_store.tests.unit import test_store_capabilities
 
 sys.modules['glance_store.common.fs_mount'] = mock.Mock()
-from glance_store._drivers import cinder  # noqa
+from glance_store._drivers import cinder # noqa
 
 
 class FakeObject(object):
